/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef OBS_PAGEALLOC_H_
#define OBS_PAGEALLOC_H_

#include <obs/obs_config.h>
#include <obs/obs_types.h>
#include <obs/obs_defines.h>
#include <obs/obs_status.h>
#include <obs/obs_region.h>

#define OBS_PA_F_EXPAND_TAIL		0x0001U
#define OBS_PA_F_EXPAND_HEAD		0x0002U
#define OBS_PA_F_SHRINK_TAIL		0x0004U
#define OBS_PA_F_SHRINK_HEAD		0x0008U
#define OBS_PA_F_ALL_FREE_ON_UNINIT	0x0010U
#define OBS_PA_F_CONCATENATE		0x0020U

typedef struct OBS_pa_private OBS_pa_private;
typedef struct OBS_pa OBS_pa;
typedef struct OBS_pa_ops OBS_pa_ops;

struct OBS_pa_ops
{
	OBS_size (*get_page_unit)(const OBS_pa_ops *);
	int (*get_alignment)(const OBS_pa_ops *);
	OBS_flags (*get_available_features)(const OBS_pa_ops *);
	OBS_status (*init)(const OBS_pa_ops *, OBS_pa *, OBS_size, OBS_flags);
	void (*uninit)(OBS_pa *);
	OBS_status (*alloc)(OBS_pa *, OBS_region *, OBS_size);
	OBS_status (*expand_head)(OBS_pa *, OBS_region *, OBS_size);
	OBS_status (*expand_tail)(OBS_pa *, OBS_region *, OBS_size);
	OBS_status (*shrink_head)(OBS_pa *, OBS_region *, OBS_size);
	OBS_status (*shrink_tail)(OBS_pa *, OBS_region *, OBS_size);
	OBS_status (*concatenate)(OBS_pa *, OBS_region *, OBS_region *);
	void (*free)(OBS_pa *, OBS_region *);
};

#define OBS_PA_MAKE_OPS_(scope, name)					      \
static OBS_size name##_get_page_unit(const OBS_pa_ops *);		      \
static int name##_get_alignment(const OBS_pa_ops *);			      \
static OBS_flags name##_get_available_features(const OBS_pa_ops *);	      \
static OBS_status name##_init(						      \
	const OBS_pa_ops *, OBS_pa *, OBS_size, OBS_flags);		      \
static void name##_uninit(OBS_pa *);					      \
static OBS_status name##_alloc(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_expand_head(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_expand_tail(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_shrink_head(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_shrink_tail(OBS_pa *, OBS_region *, OBS_size);	      \
static OBS_status name##_concatenate(OBS_pa *, OBS_region *, OBS_region *);   \
static void name##_free(OBS_pa *, OBS_region *);			      \
scope const OBS_pa_ops name##_ops =					      \
{									      \
	OBS_STRUCT_INIT_(get_page_unit, name##_get_page_unit),		      \
	OBS_STRUCT_INIT_(get_alignment, name##_get_alignment),		      \
	OBS_STRUCT_INIT_(get_available_features,			      \
			 name##_get_available_features),		      \
	OBS_STRUCT_INIT_(init, name##_init),				      \
	OBS_STRUCT_INIT_(uninit, name##_uninit),			      \
	OBS_STRUCT_INIT_(alloc, name##_alloc),				      \
	OBS_STRUCT_INIT_(expand_head, name##_expand_head),		      \
	OBS_STRUCT_INIT_(expand_tail, name##_expand_tail),		      \
	OBS_STRUCT_INIT_(shrink_head, name##_shrink_head),		      \
	OBS_STRUCT_INIT_(shrink_tail, name##_shrink_tail),		      \
	OBS_STRUCT_INIT_(concatenate, name##_concatenate),		      \
	OBS_STRUCT_INIT_(free, name##_free)				      \
}

struct OBS_pa
{
	const OBS_pa_ops *ops;
	OBS_pa_private *priv;
	OBS_size block_size;
	OBS_flags features;
};

extern const OBS_pa_ops * const obsPaDefaultOps_;
extern const OBS_pa_ops *obsPaCurrentOps_;

OBS_DEF_INLINE_FUNC_(void, obs_pa_set_current_ops, (const OBS_pa_ops *ops))
{ obsPaCurrentOps_ = ops; }

OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_current_ops, (void))
{ return obsPaCurrentOps_; }

OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_default_ops, (void))
{ return obsPaDefaultOps_; }

OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_ops_get_page_unit,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_page_unit)(ops); }

OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_get_page_unit, (void))
{ return obs_pa_ops_get_page_unit(obsPaCurrentOps_); }

OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_ops_get_available_features,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_available_features)(ops); }

OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_get_available_features, (void))
{ return obs_pa_ops_get_available_features(obsPaCurrentOps_); }

OBS_DEF_INLINE_FUNC_(int, obs_pa_ops_get_alignment,
		     (const OBS_pa_ops *ops))
{ return (*ops->get_alignment)(ops); }

OBS_DEF_INLINE_FUNC_(int, obs_pa_get_alignment, (void))
{ return obs_pa_ops_get_alignment(obsPaCurrentOps_); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_ops_init,
		     (const OBS_pa_ops *ops, OBS_pa *pa,
		      OBS_size num_units, OBS_flags features))
{ return (*ops->init)(ops, pa, num_units, features); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_init,
		     (OBS_pa *pa, OBS_size num_units, OBS_flags features))
{ return obs_pa_ops_init(obsPaCurrentOps_, pa, num_units, features); }

OBS_DEF_INLINE_FUNC_(void, obs_pa_uninit, (OBS_pa *pa))
{ (*pa->ops->uninit)(pa); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_alloc,
		     (OBS_pa *pa, OBS_region *page, OBS_size num_pages))
{ return (*pa->ops->alloc)(pa, page, num_pages); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_expand_head,
		     (OBS_pa *pa, OBS_region *page, OBS_size num_pages))
{ return (*pa->ops->expand_head)(pa, page, num_pages); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_expand_tail,
		     (OBS_pa *pa, OBS_region *page, OBS_size num_pages))
{ return (*pa->ops->expand_tail)(pa, page, num_pages); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_shrink_head,
		     (OBS_pa *pa, OBS_region *page, OBS_size num_pages))
{ return (*pa->ops->shrink_head)(pa, page, num_pages); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_shrink_tail,
		     (OBS_pa *pa, OBS_region *page, OBS_size num_pages))
{ return (*pa->ops->shrink_tail)(pa, page, num_pages); }

OBS_DEF_INLINE_FUNC_(void, obs_pa_free, (OBS_pa *pa, OBS_region *page))
{ (*pa->ops->free)(pa, page); }

OBS_DEF_INLINE_FUNC_(OBS_status, obs_pa_concatenate,
		     (OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2))
{ return (*pa->ops->concatenate)(pa, dstsrc1, src2); }

OBS_DEF_INLINE_FUNC_(OBS_size, obs_pa_get_block_size, (const OBS_pa *pa))
{ return pa->block_size; }

OBS_DEF_INLINE_FUNC_(OBS_flags, obs_pa_get_features, (const OBS_pa *pa))
{ return pa->features; }

OBS_DEF_INLINE_FUNC_(const OBS_pa_ops *, obs_pa_get_ops, (const OBS_pa *pa))
{ return pa->ops; }

#endif
