/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs.h>
#include <stdio.h>
#include <stdlib.h>

void
notice(const char *func, OBS_status errcode, int isexit)
{
	printf("%s failed: %s\n", func, obs_strstatus(errcode));
	if (isexit)
		exit(1);
}

void
dump_page(const char *msg, OBS_region *pages)
{
	printf("%s: page %p, size=%d\n", msg, pages->head, (int)pages->size);
}

int
main()
{
	OBS_pa pa;
	OBS_region pages, pages2;
	OBS_flags features;
	OBS_status ret;

	features = obs_pa_get_available_features();
	printf("features=0x%X\n", (unsigned)features);
	printf("page unit=%d\n", (int)obs_pa_get_page_unit());
	printf("alignment=%d\n", (int)obs_pa_get_alignment());
	ret = obs_pa_init(&pa, 1, features);
	if (ret)
		notice("obs_pa_init", ret, 1);

	printf("obs_pa_get_block_size: %d\n", (int)obs_pa_get_block_size(&pa));

	ret = obs_pa_alloc(&pa, &pages, 2);
	if (ret)
		notice("obs_pa_alloc", ret, 1);
	dump_page("alloc", &pages);
	ret = obs_pa_expand_tail(&pa, &pages, 1);
	if (ret)
		notice("obs_pa_expand_tail", ret,
		       features & OBS_PA_F_EXPAND_TAIL);
	dump_page("expand_tail(1)", &pages);
	ret = obs_pa_shrink_tail(&pa, &pages, 1);
	if (ret)
		notice("obs_pa_shrink_tail", ret,
		       features & OBS_PA_F_SHRINK_TAIL);
	dump_page("shrink_tail(1)", &pages);
	ret = obs_pa_alloc(&pa, &pages2, 1);
	if (ret)
		notice("obs_pa_alloc(pages2)", ret, 1);
	dump_page("alloc(pages2)", &pages2);
	ret = obs_pa_concatenate(&pa, &pages, &pages2);
	if (ret)
		notice("obs_pa_concatenate", ret, 0);
	dump_page("concat", &pages);
	obs_pa_free(&pa, &pages);
	dump_page("free", &pages);
	obs_pa_free(&pa, &pages2);
	dump_page("free", &pages2);

	obs_pa_uninit(&pa);

	return 0;
}
