/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs.h>
#include <stdio.h>
#include <stdlib.h>

const OBS_pa_ops *curops;

OBS_PA_MAKE_OPS_(static, wrapped);

OBS_size
/*ARGSUSED*/
wrapped_get_page_unit(const OBS_pa_ops *ops)
{
	OBS_size ret = obs_pa_ops_get_page_unit(curops);
	printf("  get_page_unit=%u\n", (unsigned)curops);
	return ret;
}

int
/*ARGSUSED*/
wrapped_get_alignment(const OBS_pa_ops *ops)
{
	int ret = obs_pa_ops_get_page_unit(curops);
	printf("  get_alignment=%d\n", ret);
	return ret;
}

OBS_flags
/*ARGSUSED*/
wrapped_get_available_features(const OBS_pa_ops *ops)
{
	OBS_flags ret = obs_pa_ops_get_available_features(curops);
	printf("  get_available_features=0x%X\n", (unsigned)ret);
	return ret;
}

OBS_status
/*ARGSUSED*/
wrapped_init(const OBS_pa_ops *ops, OBS_pa *pa, OBS_size num_units,
	     OBS_flags features)
{
	OBS_status ret;

	printf("  init(num_units=%u, features=0x%X) = ",
	       (unsigned)num_units, (unsigned)features);
	ret = obs_pa_ops_init(curops, pa, num_units, features);
	pa->ops = &wrapped_ops;
	printf("%s\n", obs_strstatus(ret));

	return ret;
}

void wrapped_uninit(OBS_pa *pa)
{
	printf("  uninit\n");
	(*curops->uninit)(pa);
}

OBS_status
wrapped_alloc(OBS_pa *pa, OBS_region *pages, OBS_size num_pages)
{
	OBS_status ret;

	printf("  alloc(num_pages=%u) = ", (unsigned)num_pages);
	ret = (*curops->alloc)(pa, pages, num_pages);
	printf("%s (%p,%u)\n", obs_strstatus(ret),
	       pages->head, (unsigned)pages->size);

	return ret;
}

OBS_status
/*ARGSUSED*/
wrapped_expand_head(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{
	OBS_status ret;

	printf("  expand_head(%p,%u),incr=%u) = ",
	       pages->head, (unsigned)pages->size, (unsigned)incr);
	ret = (*curops->expand_head)(pa, pages, incr);
	printf("%s (%p,%u)\n", obs_strstatus(ret),
	       pages->head, (unsigned)pages->size);

	return ret;
}

OBS_status
/*ARGSUSED*/
wrapped_expand_tail(OBS_pa *pa, OBS_region *pages, OBS_size incr)
{
	OBS_status ret;

	printf("  expand_tail(%p,%u),incr=%u) = ",
	       pages->head, (unsigned)pages->size, (unsigned)incr);
	ret = (*curops->expand_tail)(pa, pages, incr);
	printf("%s (%p,%u)\n", obs_strstatus(ret),
	       pages->head, (unsigned)pages->size);

	return ret;
}

OBS_status
wrapped_shrink_head(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{
	OBS_status ret;

	printf("  shrink_head(%p,%u),decr=%u) = ",
	       pages->head, (unsigned)pages->size, (unsigned)decr);
	ret = (*curops->shrink_head)(pa, pages, decr);
	printf("%s (%p,%u)\n", obs_strstatus(ret),
	       pages->head, (unsigned)pages->size);

	return ret;
}

OBS_status
wrapped_shrink_tail(OBS_pa *pa, OBS_region *pages, OBS_size decr)
{
	OBS_status ret;

	printf("  shrink_tail(%p,%u),decr=%u) = ",
	       pages->head, (unsigned)pages->size, (unsigned)decr);
	ret = (*curops->shrink_tail)(pa, pages, decr);
	printf("%s (%p,%u)\n", obs_strstatus(ret),
	       pages->head, (unsigned)pages->size);

	return ret;
}

OBS_status
/*ARGSUSED*/
wrapped_concatenate(OBS_pa *pa, OBS_region *dstsrc1, OBS_region *src2)
{
	OBS_status ret;

	printf("  concatenate((%p,%u), (%p,%u)) = ",
	       dstsrc1->head, (unsigned)dstsrc1->size,
	       src2->head, (unsigned)src2->size);
	ret = (*curops->concatenate)(pa, dstsrc1, src2);
	printf("%s (%p,%u), (%p,%u)\n", obs_strstatus(ret),
	       dstsrc1->head, (unsigned)dstsrc1->size,
	       src2->head, (unsigned)src2->size);

	return ret;
}

void
/*ARGSUSED*/
wrapped_free(OBS_pa *pa, OBS_region *pages)
{
	printf("  free (%p,%u)\n", pages->head, (unsigned)pages->size);
	(*curops->free)(pa, pages);
}


void
notice(const char *func, OBS_status errcode, int isexit)
{
	printf("%s failed: %s\n", func, obs_strstatus(errcode));
	if (isexit)
		exit(1);
}

void
dump_reg(const char *mes, const OBS_region *reg)
{
	printf("%s: ", mes);
	if (obs_region_is_valid(reg))
		printf("%p (%d)\n", reg->head, reg->size);
	else
		printf("invalid\n");
}

int
main()
{
	OBS_status ret;
	OBS_stack obs;
	OBS_region reg1;

	curops = obs_pa_get_current_ops();
	obs_pa_set_current_ops(&wrapped_ops);

	ret = obs_stack_init(&obs, 1, 0, OBS_ALIGN_4);
	if (ret)
		notice("obs_stack_init", ret, 1);

	ret = obs_stack_alloc(&obs, 10);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	reg1 = *obs_stack_get_top(&obs);
	dump_reg("alloc(10)", obs_stack_get_top(&obs));
	ret = obs_stack_alloc(&obs, 20);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(20)", obs_stack_get_top(&obs));
	ret = obs_stack_truncate_after(&obs, reg1.head);
	if (ret)
		notice("obs_stack_truncate_after", ret, 1);
	dump_reg("trunc", obs_stack_get_top(&obs));
	ret = obs_stack_alloc(&obs, 4000);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(4000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize_top(&obs, 8000);
	if (ret)
		notice("obs_stack_resize_top", ret, 1);
	dump_reg("resize(8000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize_top(&obs, 12000);
	if (ret)
		notice("obs_stack_resize_top", ret, 1);
	dump_reg("resize(12000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize_top(&obs, 16000);
	if (ret)
		notice("obs_stack_resize_top", ret, 1);
	dump_reg("resize(16000)", obs_stack_get_top(&obs));
	reg1 = *obs_stack_get_top(&obs);
	ret = obs_stack_alloc(&obs, 16000);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(16000)", obs_stack_get_top(&obs));

	ret = obs_stack_truncate_before(
		&obs, (void *)((char *)reg1.head+16000));
	if (ret)
		notice("obs_stack_truncate_before", ret, 1);
	dump_reg("trunc_before", obs_stack_get_top(&obs));

	obs_stack_uninit(&obs);

	return 0;
}
